// Cheap DSP dsPICradio board LCD routines
// This is polling driver with 8-bit interface

// Juha Niinikoski OH2NLT 26.08.2005

// Modified for Digital RX hardware 12.09.2006
// MPLAB C30 version, OH2GWE, 2006.10.17

// IRQ mask added to LCD I/O port operations. IRQ tone output needs this protection.
// S-meter roll over corrected, interface changed to int 02.12.2006
// LCD HEX output added 09.07.2007

// Includes
#include <stdio.h>
#include <string.h>
#include "juma-trx2.h"
#include "trx2_eeprom.h"			// EEPROM storage structure

// External Functions
extern void us_delay(unsigned int);
extern void ms_delay(unsigned int);

// External Variables
extern char lcdpbuff[];				// LCD print buffer, longer than the display for prints
extern int PTT_IN;
extern int band_in_use;
extern unsigned int ms_counter;

// Local variables
char scale_dots[] = {				// 4 vertical dashes, S9 marker, 3 vertical dashes (Used by draw_s_meter())
						0x00,
						0x00,
						0x00,
						0x00,
						0x00,
						0x00,
						0x00,
						0x00
						};

// Local Data
const char init_lcd[] = {			// LCD initialisation data
						0x38,		// 8-bit, 2 rows, 5x7 matrix, display on		
						0x38,
						0x38,
						0x0C,		// Cursor off, no blink
						0x01,
						0x06		// Increment, no display shift
						};

const char *scale_fmt[] = {
						"%6.1fW",
						"SWR%4.1f",
						"%6.2fV",
						"%6.2fA"
						};

const double scale_factor[] = {
						1000000000.0,
						100.0,
						1000000.0,
						2000000.0
						};

const char RIT_Fmt[] = {"R%+5.2f"};

// User Configuration Data
extern union
	{
	struct defval defval;
	unsigned int storage[sizeof(struct defval) / 2];
	} eeprom;

void clk_lcd(unsigned char ldata)			// Set data & clock both LCD controllers
	{
	__asm__ volatile ("disi #9");			// Protect IRQ tone output to RD1 I/O
	LCD_DATA = (LCD_DATA & ~LCD_MASK) | ((ldata << LCD_SHIFT) & LCD_MASK);	// set LCD data lines
	us_delay(1);							// Extra delay for filter board latch wires
	LCD_E = 1;
	us_delay(1);
	LCD_E = 0;
	us_delay(1); 							// Test
	}

void wait_lcd_rdy(void)						// Wait until LCD module is ready
	{
	unsigned int BF_AC;						// Busy Flag & address counter

	BF_AC = (0x80 << LCD_SHIFT);			// Set busy for first round
	__asm__ volatile ("disi #4");			// Protect IRQ tone output to RD1 I/O
	LCD_TRIS |= LCD_MASK;					// Make data port input, LCD bus direction bits = 1
	LCD_RW = 1;								// Read
	LCD_RS = 0;								// Command mode

	while(BF_AC & (0x80 << LCD_SHIFT))
		{
		us_delay(1);						// 18.01.2004 for slow display
		LCD_E = 1;
		us_delay(2);
		BF_AC = LCD_BUS & LCD_MASK;			// Read from port pins
		LCD_E = 0;
		}

	LCD_RW = 0;
	LCD_RS = 1;
	__asm__ volatile ("disi #4");			// Protect IRQ tone output to RD1 I/O
	LCD_TRIS &= (~LCD_MASK);				// Return to Write mode
	}

void lcd_cmd(unsigned char cmnd)
	{
	wait_lcd_rdy();
	LCD_RS = 0;								// Switch to command mode		
	clk_lcd(cmnd);							// Set cursor command + cursor position
	LCD_RS = 1;								// Switch back to data mode
	}

void clear_lcd(void)						// Clear selected display
	{
	lcd_cmd(CLEAR);							// If you replace clear_lcd() with this command, you will use more memory.
	}

void initlcd(void)							// Init selected LCD controller
	{
	int i;

	LCD_RS = 0;								// Set command mode
	LCD_RW = 0;
	ms_delay(15);							// Start delay

	for(i = 0; i < 6; i++)
		{
		clk_lcd(init_lcd[i]);				// Write init data with delays
		ms_delay(5);
		}

	LCD_RS = 1;								// Set data mode
	set_pwm4_dac(DEFAULT_CONTRAST);			// Initialise the default constrast setting
	set_pwm3_dac(DEFAULT_BL);				// and default backlighting.
	}

void lcdoutch(unsigned char lcd_char)		// Output character to selected display
	{
	wait_lcd_rdy();
	clk_lcd(lcd_char);
	}

// HD44780 character generator routines
// Modified for SOLOMON LM1125SYLU1 display
void set_ch_bits(char data, char count)		// Set chgen bits. Data, number of sets
	{
	while(count--) lcdoutch(data);
	}

void set_chgen(int type)					// Write bar symbols to HD44780 character generator RAM
	{
	lcd_cmd(0x00 | ADRSET);

	switch(type)
		{
		case 0:		//	Original Scale
//	CGRAM 00 - Empty Block and part = 0
			set_ch_bits(0x00, 2);	// .....
									// .....
			set_ch_bits(0x10, 3);	// @....
									// @....
									// @....
			set_ch_bits(0x00, 3);	// .....
									// .....
									// .....

//	CGRAM 01 - part = 1
			set_ch_bits(0x00, 1);	// .....
			set_ch_bits(0x10, 5);	// @....
									// @....
									// @....
									// @....
									// @....
			set_ch_bits(0x00, 2);	// .....
									// .....

//	CGRAM 02 - part = 2
			set_ch_bits(0x00, 1);	// .....
			set_ch_bits(0x10, 5);	// @....
									// @....
									// @....
									// @....
									// @....
			set_ch_bits(0x00, 2);	// .....
									// .....

//	CGRAM 03 - part = 3
			set_ch_bits(0x00, 1);	// .....
			set_ch_bits(0x14, 5);	// @.@..
									// @.@..
									// @.@..
									// @.@..
									// @.@..
			set_ch_bits(0x00, 2);	// .....
									// .....


//	CGRAM 04 - part = 4
			set_ch_bits(0x00, 1);	// .....
			set_ch_bits(0x14, 5);	// @.@..
									// @.@..
									// @.@..
									// @.@..
									// @.@..
			set_ch_bits(0x00, 2);	// .....
									// .....

// CGRAM 05 - Full Block and part = 5
			set_ch_bits(0x00, 1);	// .....
			set_ch_bits(0x15, 5);	// @.@.@
									// @.@.@
									// @.@.@
									// @.@.@
									// @.@.@
			set_ch_bits(0x00, 2);	// .....
									// .....

// CGRAM 06 - S-Meter S9 Marker
			set_ch_bits(0x00, 2);	// .....
									// .....
			set_ch_bits(0x10, 1);	// @....
			set_ch_bits(0x00, 1);	// .....
			set_ch_bits(0x10, 1);	// @....
			set_ch_bits(0x00, 3);	// .....
									// .....
									// .....

// CGRAM 07 - End-Of-Scale
			set_ch_bits(0x00, 2);	// .....
									// .....
			set_ch_bits(0x10, 3);	// @....
									// @....
									// @....
			set_ch_bits(0x00, 3);	// .....
									// .....
									// .....

			scale_dots[4] = 0x06;	// Special S9 marker for Original scale
		break;

		case 1:		// Large Scale Markers
// CGRAM 00 - Empty block, and part = 0
			set_ch_bits(0x00, 1);	// .....
			set_ch_bits(0x10, 2);	// @....
									// @....
			set_ch_bits(0x15, 2);	// @.@.@
									// @.@.@
			set_ch_bits(0x10, 2);	// @....
									// @....
			set_ch_bits(0x00, 1);	// .....

// CGRAM 01 (part = 1)
			set_ch_bits(0x00, 1);	// .....
			set_ch_bits(0x10, 1);	// @....
			set_ch_bits(0x18, 1);	// @@...
			set_ch_bits(0x1D, 2);	// @@@.@
									// @@@.@
			set_ch_bits(0x18, 1);	// @@...
			set_ch_bits(0x10, 1);	// @....
			set_ch_bits(0x00, 1);	// .....

// CGRAM 02 (part = 2)
			set_ch_bits(0x00, 1);	// .....
			set_ch_bits(0x10, 1);	// @....
			set_ch_bits(0x1C, 1);	// @@@..
			set_ch_bits(0x1D, 2);	// @@@.@
									// @@@.@
			set_ch_bits(0x1C, 1);	// @@@..
			set_ch_bits(0x10, 1);	// @....
			set_ch_bits(0x00, 1);	// .....

// CGRAM 03	(part = 3)
			set_ch_bits(0x00, 1);	// .....
			set_ch_bits(0x10, 1);	// @....
			set_ch_bits(0x1E, 1);	// @@@@.
			set_ch_bits(0x1F, 2);	// @@@@@
									// @@@@@
			set_ch_bits(0x1E, 1);	// @@@@.
			set_ch_bits(0x10, 1);	// @....
			set_ch_bits(0x00, 1);	// .....

// CGRAM 04 (part = 4)
			set_ch_bits(0x00, 1);	// .....
			set_ch_bits(0x10, 1);	// @....
			set_ch_bits(0x1F, 4);	// @@@@@
									// @@@@@
									// @@@@@
									// @@@@@
			set_ch_bits(0x10, 1);	// @....
			set_ch_bits(0x00, 1);	// .....

// CGRAM 05 - (part = 5) Fully-Filled Block
			set_ch_bits(0x00, 1);	// .....
			set_ch_bits(0x10, 1);	// @....
			set_ch_bits(0x1F, 4);	// @@@@@
									// @@@@@
									// @@@@@
									// @@@@@
			set_ch_bits(0x10, 1);	// @....
			set_ch_bits(0x00, 1);	// .....

// CGRAM 06 - End-Of-Scale Marker
			set_ch_bits(0x00, 1);	// .....
			set_ch_bits(0x10, 6);	// @....
									// @....
									// @....
									// @....
									// @....
									// @....
			set_ch_bits(0x00, 1);	// .....

// CGRAM 07 - End-Of-Scale Marker
			set_ch_bits(0x00, 1);	// .....
			set_ch_bits(0x10, 6);	// @....
									// @....
									// @....
									// @....
									// @....
									// @....
			set_ch_bits(0x00, 1);	// .....

			scale_dots[4] = 0x00;	// Normal 'empty' block for Large Scale
		break;

		case 2:		// Small scale markers
// CGRAM 00 - Empty block, and part = 0
			set_ch_bits(0x00, 2);	// .....
									// .....
			set_ch_bits(0x10, 1);	// @....
			set_ch_bits(0x15, 2);	// @.@.@
									// @.@.@
			set_ch_bits(0x10, 1);	// @....
			set_ch_bits(0x00, 2);	// .....
									// .....

// CGRAM 01 (part = 1)
			set_ch_bits(0x00, 2);	// .....
									// .....
			set_ch_bits(0x18, 1);	// @@...
			set_ch_bits(0x1D, 2);	// @@@.@
									// @@@.@
			set_ch_bits(0x18, 1);	// @@...
			set_ch_bits(0x00, 2);	// .....
									// .....

// CGRAM 02 (part = 2)
			set_ch_bits(0x00, 2);	// .....
									// .....
			set_ch_bits(0x1C, 1);	// @@@..
			set_ch_bits(0x1D, 2);	// @@@.@
									// @@@.@
			set_ch_bits(0x1C, 1);	// @@@..
			set_ch_bits(0x00, 2);	// .....
									// .....

// CGRAM 03	(part = 3)
			set_ch_bits(0x00, 2);	// .....
									// .....
			set_ch_bits(0x1E, 1);	// @@@@.
			set_ch_bits(0x1F, 2);	// @@@@@
									// @@@@@
			set_ch_bits(0x1E, 1);	// @@@@.
			set_ch_bits(0x00, 2);	// .....
									// .....

// CGRAM 04 (part = 4)
			set_ch_bits(0x00, 2);	// .....
									// .....
			set_ch_bits(0x1F, 4);	// @@@@@
									// @@@@@
									// @@@@@
									// @@@@@
			set_ch_bits(0x00, 2);	// .....
									// .....

// CGRAM 05 - (part = 5) Fully-Filled Block
			set_ch_bits(0x00, 2);	// .....
									// .....
			set_ch_bits(0x1F, 4);	// @@@@@
									// @@@@@
									// @@@@@
									// @@@@@
			set_ch_bits(0x00, 2);	// .....
									// .....

// CGRAM 06 - End-Of-Scale Marker
			set_ch_bits(0x00, 2);	// .....
									// .....
			set_ch_bits(0x10, 4);	// @....
									// @....
									// @....
									// @....
			set_ch_bits(0x00, 2);	// .....
									// .....
// CGRAM 07 - End-Of-Scale Marker
			set_ch_bits(0x00, 2);	// .....
									// .....
			set_ch_bits(0x10, 4);	// @....
									// @....
									// @....
									// @....
			set_ch_bits(0x00, 2);	// .....
									// .....

			scale_dots[4] = 0x00;	// Normal 'empty' block for Small Scale
		break;
		}

	clear_lcd();					// Back to normal mode (clear LCD)
	}

// Draw 16 long S-meter, scale = 0...96 (48 visible steps)
// Draw 8 long S-meter, scale = 0...48 (24 visible steps)
void draw_s_meter(int len)
	{
	int i, whole, part;
	char c;
	
	if(len > MAX_GRAPH) len = MAX_GRAPH;	// Cut to max MAX_GRAPH = 48 (Maximum graph value)

	whole = len / FONT_W;					// Calculate graph. FONT_W = 6 whole = fully filled blocks
	part = len % FONT_W;					// part = partially filled block, index into scale_dots[] array

	for(i = 0; i < DISP_LEN; i++)
		{
		if(i < whole) c = FULL_BLOCK;		// Fully-filled block
		else if(part)						// If there is a remainder,
				{
				c = part;					// part is index into partial block character map
				part = 0;					// Clear part so that on next round, we display empty blocks.
				}
				else c = scale_dots[i];		// If no remainder, then simply empty blocks. If the scale is set to Original
											// then the S9 block is special, and is already in the scale_dots[] array.
		lcdoutch(c);
		}
// Using PPT_IN_IO fixes Voice Memory TX power meter missing end-of-scale bug when S-Meter is set to Numeric. A.Ryan - 5B4AIY - 20/AUG/2013
// PTT_IN_IO = 1 for RX, eeprom.defval.S_Meter_Type = 1 for Numeric
	lcdoutch((PTT_IN_IO && eeprom.defval.S_Meter_Type) ? ' ' : END_OF_SCALE);
	}

// General I/O functions
// Display null terminated string
void lcd_putst(register const char *str)
	{
	while(*str) lcdoutch(*str++);			// Much more concise form. - 5B4AIY
	}

long adjust_frequency(long freq, long divisor)
	{
	long f;

	f = freq / divisor;
	lcdoutch((char)(f + '0'));
	return freq - (f * divisor);
	}

void display_out_of_band(void)
	{
	char c = '.';

	if(!band_in_use && (ms_counter > 500)) c = ' ';

	lcdoutch(c);
	}

long Mega_Hz(long freq)
	{
	if(freq > 9999999L) freq = adjust_frequency(freq, 10000000L);		// Tens of MHz
	else lcdoutch(' ');

	return adjust_frequency(freq, 1000000L);			// Units of MHz
	}

long hundred_ten_kilo(long freq)
	{
	freq = adjust_frequency(freq, 100000L);				// Hundred Kilohertz
	freq = adjust_frequency(freq, 10000L);				// Ten Kilohertz
	return adjust_frequency(freq, 1000L);				// Kilohertz
	}

long hundred_hz_ten_hz(long freq)
	{
	freq = adjust_frequency(freq, 100L);				// Hundred Hertz
	return adjust_frequency(freq, 10L);					// Ten Hertz
	}

long mega_dot_hundred_ten(long freq)
	{
	freq = Mega_Hz(freq);								// Tens & Units of MHz
	display_out_of_band();								// Blink decimal point if frequency is out-of-band
	freq = hundred_ten_kilo(freq);						// Hundreds, Tens & Units of KHz
	return hundred_hz_ten_hz(freq);						// Hundred Hz & Tens Hz
	}

long hundred_ten_kilo_dot_hundred_ten_hz(long freq)
	{
	freq = hundred_ten_kilo(freq);						// Hundreds, Tens & Units of KHz
	display_out_of_band();								// Blink decimal point if frequency is out-of-band
	return hundred_hz_ten_hz(freq);						// Hundred Hz & Tens Hz
	}

long kilo_dot_hundred(long freq)
	{
	freq = adjust_frequency(freq, 1000L);				// Kilohertz
	display_out_of_band();								// Blink decimal point if frequency is out-of-band
	return hundred_hz_ten_hz(freq);						// Hundred Hz & Tens Hz
	}
/*
  Display frequency in decimal format - new version - 5B4AIY
  The original called disp_long() and passed two parameters, x, and mode, but disp_long() was never
  called anywhere else in the program, and thus it made no sense to do two function calls where only
  one was required.
  Final optimisation, collected all common code into functions, and removed repetitive code.
  This saved 852 bytes of code. A. Ryan - 5B4AIY 21/JUN/2012
  More optimisation, removed original step logic, now no longer required, and simplified
  the trailing character logic. This saved a further 235 bytes of code. A.Ryan - 5B4AIY - 13/JUL/12
*/
void disp_freq(long freq)
	{
	char c = ' ';												// Setup the final character
/*
  Assume the input frequency is: 23.456789 MHz
  New Display, Fixed-A 1Hz: 23.456789
  If the frequency is less than 1MHz, then the display is: 999.999, with leading blanks rather than zeros
  The low frequency limit is 0, so for frequencies below 1kHz the display will be: 999, or 99, or 9, or 0,
  with leading blanks.
*/
	switch(eeprom.defval.Freq_Display)
		{
		case ORIGINAL:											// Original Display - 23.456.79
			freq = Mega_Hz(freq);								// 10Hz resolution, two decimal points, leading zeros
			lcdoutch('.');
			freq = hundred_ten_kilo_dot_hundred_ten_hz(freq);
			return;
		break;

		case NEW_DISP:											// New Display - 23.456789
			c = (mega_dot_hundred_ten(freq) + '0');				// 1Hz resolution, leading zeros
		break;
/*
  Do not be tempted to change the format to: sprintf(lcdpbuff, "%9.6f", ((double)freq) / 1000000.0);
  Although this will work for low frequencies, above about 10MHz and especially above 20MHz there will
  be display anomalies in the Hz digit. This is due to rounding errors accumulating in the conversion
  from long integer to double, and are simply a limitation of the implementation of sprintf() on this
  compiler. Did you really think I wouldn't try the obvious first? For the 10Hz and coarser resolutions
  the conversion is done correctly, and no display anomalies occur. - 5B4AIY 20/SEP/11
  Unfortunately, using sprintf() and then using a loop to move characters in the buffer took a long time,
  the loop cycle time increased from 7mS to 14mS when displaying frequencies to 1Hz! The repeated division
  version was much faster, adding very little time to the loop, the loop time is then 6.3mSec. Using
  the other versions, (case 1 through 6) the loop cycle time is 7mSec.
  Using sprintf() and memmove() to move the characters increased the loop time to 14mSec again.
  By using a temporary variable, f, and only performing 1 division, the loop cycle time improved to 5.8mSec.
  Building the buffer rather than simply putting characters on the LCD increased the time to 5.9mSec.
  Using the Old display, the loop time is 6.9mSec.
  Case 0 now has a loop cycle time of 5.7mSec
  Using the Old frequency display but the new display algorithms the loop cycle time improved from 6.9mS
  to 5.7mS
  A. Ryan - 5B4AIY 23/OCT/2011
  After numerous re-writes, this turns out to be the most efficient way of achieving this display. Re-writing
  adjust_frequency() to test for a zero and print a space if a flag is set simply complicates this function
  with no contributory benefit to the other displays. Even though it is laborious to have to test each decade,
  this is still the shortest and most efficient way of accomplishing this. A. Ryan 16/JUL/2012

  Note: lcd_spc() is a macro definition in juma_trx2.h A. Ryan 31/JUL/2012
*/
		case FIXED_A:												// Fixed-A - 23.456789, 1Hz resolution, leading blanks
			if(freq > 999999L) freq = mega_dot_hundred_ten(freq);	// 10MHz/1MHz
			else if(freq > 99999L)									// 100kHz
					{
					lcd_spc(2);
					freq = hundred_ten_kilo_dot_hundred_ten_hz(freq);
					}
				else if(freq > 9999L)								// 10kHz
					{
					lcd_spc(3);
					freq = adjust_frequency(freq, 10000L);
					freq = kilo_dot_hundred(freq);
					}
					else if(freq > 999L)							// 1kHz
						{
						lcd_spc(4);
						freq = kilo_dot_hundred(freq);
						}
						else if(freq > 99L)							// 100Hz
							{
							lcd_spc(6);
							freq = hundred_hz_ten_hz(freq);
							}
							else if(freq > 9L)						// 10Hz
								{
								lcd_spc(7);
								freq = adjust_frequency(freq, 10L);
								}
								else lcd_spc(8);

			c = (char)freq + '0';									// 1Hz
		break;

		case FIXED_B:												// Fixed-B - 23.45679
			freq = mega_dot_hundred_ten(freq);						// 10Hz resolution, leading zeros
		break;
		}

	lcdoutch(c);													// Trailing frequency character or space 
	}
/*
  Display RIT format - new version 5B4AIY
  The display is: R+1.00 or R-1.00 or R 0.00 (Maximum offset is 1,000Hz)
  With this implementation, adjusting the offset to exactly 0Hz will display: R 0.00
  with a +1Hz offset the display will be: R+0.00, with a -1Hz offset the display will be: R-0.00,
  thus the user can set the RIT frequency to exactly zero by adjusting for no sign.
*/
void disp_rit(int x)
	{
	sprintf(lcdpbuff, RIT_Fmt, ((double)x) / 1000.0);	// The + in the format statement forces a leading +/- symbol

	if(x == 0) lcdpbuff[1] = ' ';						// If the offset is exactly zero, do not display a sign.

	lcd_putst(lcdpbuff);
}

// Display meter format - new version 5B4AIY
// Replaced SWITCH/CASE construct with a more efficient indexed array construct. A.Ryan - 5B4AIY - 12/JAN/2014
void disp_meter(double value, int index)
	{
	sprintf(lcdpbuff, scale_fmt[index], value / scale_factor[index]);
	lcd_putst(lcdpbuff);
	}

