// JUMA-TRX2 service & setup routines
// Juha Niinikoski, OH2NLT

// Service routines moved to separate file service.c 27.12.2006
// Some displays changed, TX can operate, last power off VFO & freq in use
// Minor text corrections 30.10.2008
// Revised display prompts - 5B4AIY
// Added adjusment limit checks - 5B4AIY
// Added Overcurrent, Over Voltage, and Under Voltage trips - 5B4AIY
// Moved Band Limit setting from User Configuration module to here - 5B4AIY
// Added High SWR Alarm adjustable delay. 5B4AIY

#include <stdio.h>							// Added - 5B4AIY
#include "juma-trx2.h"						// Project definitions, macros and constants
#include "trx2_eeprom.h"					// EEPROM structure definitions

#define MAX_SERVICE_PAGES	15				// Number of display pages

// External References
// External Functions
extern void beep(int, int);								// Generate beep, period, length
extern void ms_delay( unsigned int);					// Delay
extern int encoder_get(void);							// Read encoder
extern void reset_encoder(void);						// Reset encoder
extern void clear_lcd(void);							// Clear LCD
extern void lcd_cmd(unsigned char);
extern void lcdoutch(unsigned char);					// Display single character
extern void lcd_putst(register const char *);			// LCD put string
extern void disp_meter(double, int);					// Display meters
extern void display_s_meter(int);						// Display scaled S-meter
extern void disp_fwd_pwr(void);							// Display wattmeter
extern void disp_id(void);								// Display RF Amp Drain current
extern void get_pwr(void);								// Get RF power data
extern void send_packet_to_74hc595(void);				// Main & PA board I/O
extern int convert_adc12(unsigned int);					// Added - 5B4AIY
extern int encoder_get_coarse(int);						// Added - 5B4AIY
extern void set_factory_defaults(void);					// Added - 5B4AIY
extern void save_defaults(void);						// Added - 5B4AIY
extern void save_calval(int);							// Added - 5B4AIY
extern int save_settings(int, int);						// Added - 5B4AIY
extern void get_value(int *, int, int, int);			// Added - 5B4AIY
extern void set_decade_value(int *, int, int);			// Added - 5B4AIY
extern void set_chgen(int);
extern void dds_load(unsigned long);					// Added - 5B4AIY
extern void max_min(int *, int, int);
extern void warning_beep(void);

// External Data
extern int fwd_adval;				// Forward & reverse power raw values, measured only when TX
extern int rev_adval;
extern long fwd_pwr;				// Relative forward power
extern long rev_pwr;				// Relative reverse power
extern long out_pwr;				// Scaled output power = (fwd_pwr - rev_pwr) * scale
extern int id_raw_value;			// Raw drain current
extern int PTT_IN;					// PTT switch value
extern int tune_on;					// Tune state flag
extern char lcdpbuff[];				// Standard LCD display buffer
extern unsigned int long_push;		// Long push timer for buttons
extern char *band_limits[];
extern char *on_off[];
extern char spc_or_bar[];
extern char _16s[];
extern int s_meter_const;			// Calculated S-Meter Calibration Constant
extern long freq_in_use;
extern int vfo_idx;
extern int latch;
extern int decay_counter;

// Local Data
const char cal_factor[] = {"Cal Factor: %4i"};
const char trip_frmt[] = {"Trip:%11s"};
const char trip_msg[] = {"Limit:%9.2f%c"};
const char Restore_Defaults[] = {"\n\rRestore Factory Defaults: %s\n\r"};
const char *service_prompt[] = {						// LCD display prompt messages (Line 1)
								"Set Ref Osc Freq",		// 0
								"Supply:  ",			// 1
								"RF Amp:  ",			// 2
								"Forward: ",			// 3
								"",						// 4 - S-meter display is on the first line, this is a place-holder
								"Beep Time  0=Off",		// 5
								"High SWR Trip   ",		// 6
								"SWR Alarm Delay ",		// 7
								"Overcurrent     ",		// 8
								"Overcurrent Trip",		// 9
								"Overvoltage     ",		// 10
								"Overvoltage Trip",		// 11
								"Low Voltage     ",		// 12
								"Low Voltage Trip",		// 13
								"Set Band Limits ",		// 14
								"  FAST/VFAST to "		// 15
								};
const char *settings_saved[] = {
								"Existing Calibration settings re-loaded from",
								"New Calibration settings saved to"
								};
const char *cancelled_restored[] = {
								"Cancelled!",
								"Restored!"
								};
const char svc_0_msg[] = {"Osc:%9li Hz"};
const char svc_4_msg[] = {"S-Meter"};
const char svc_5_msg1[] = {"Beep:%8i mS"};
const char svc_5_msg2[] = {"Beep:%11s"};
const char svc_6_msg[] = {"Limit:%8.1f:1"};
const char svc_7_msg1[] = {"Time:%6.1f Secs"};
const char svc_7_msg3[] = {"Disabled"};
const char svc_15_msg[] = {"restore defaults"};
const char Display_Next_msg[] = {"\n\rService & Calibration Mode\n\rDISPLAY/CONFIG = Parameter Select\n\rFAST/VFAST = Save & Exit\n\r"};
const char EEPROM_Saved[] = {"\n\r%s EEPROM\n\r"};

int i, cal_page = 0;

// User Configuration Data
extern union
	{
	struct defval defval;
	} eeprom;

// System Calibration Values
extern union
	{
	struct calval calval;
	} cal;

// Service & calibration functions
void display_alarm_msg(int flag)
	{
	sprintf(lcdpbuff, trip_frmt, (flag) ? on_off[ON] : on_off[OFF]);
	}

long set_range(long value, int rate, long cal_factor, unsigned long min, unsigned long max)
	{
	if(latch == FALSE)		// If Fast Page Select is not active, adjust value, otherwise skip
		value += (rate) ? encoder_get_coarse(rate) : encoder_get();

	value *= cal_factor;
	value = (value > max) ? max : (value < min) ? min : value;
	return value / cal_factor;
	}

void set_flag(int enable, int disable)
	{
	int i;

	if(latch) return;		// If Fast Page Select is active, skip

	i = encoder_get_coarse(ENCODER_SLOW);

	if(i > 0)								// Enable
		cal.calval.alarm_flags |= enable;

	if(i < 0)								// Disable
		cal.calval.alarm_flags &= disable;
	}

void svc_0(void)							// Set Master Clock calibration factor
	{
	cal.calval.ref_osc = set_range(cal.calval.ref_osc, SLOW, 1, MIN_OSC, MAX_OSC);
	sprintf((char *)lcdpbuff, svc_0_msg, cal.calval.ref_osc);	// Added cast - 5B4AIY
	}

void svc_1(void)							// Set battery voltage calibration factor
	{
	double batt;

	batt = (double)convert_adc12(BATT_CH);
	batt *= (double)cal.calval.batt_mult;
	disp_meter(batt, VOLTS);										// New version 5B4AIY
	get_value(&cal.calval.batt_mult, ENCODER_NORMAL, BATT_MULT_MIN, BATT_MULT_MAX);
	sprintf((char *)lcdpbuff, cal_factor, cal.calval.batt_mult);	// Show scale multiplier - Added cast - 5B4AIY
	}

void svc_2(void)							// Set drain current calibration factor
	{
	id_raw_value = convert_adc12(ID_CUR);
	disp_id();														// Display drain current
	get_value(&cal.calval.id_mult, ENCODER_NORMAL, ID_MULT_MIN, ID_MULT_MAX);
	sprintf((char *)lcdpbuff, cal_factor, cal.calval.id_mult);		// Added cast - 5B4AIY
	}

void svc_3(void)							// Set power meter calibration factor
	{
	get_pwr();
	disp_fwd_pwr();													// Display Wattmeter
	get_value(&cal.calval.fwd_pwr_mult, ENCODER_NORMAL, FWD_PWR_MULT_MIN, FWD_PWR_MULT_MAX);
	sprintf((char *)lcdpbuff, cal_factor, cal.calval.fwd_pwr_mult);	// Added cast - 5B4AIY
	}
/*
  The S-meter is designed with a DC output voltage of 1V for S9, and 2V for S9+40dB. A 1V input to the A-D
  convertor results in a count of 820. Therefore the scaling factor to read half-scale is 820 / 24 = 34.
  The scale is a 48-point scale, and for a 2V input with a count of 1640 and a scale factor of 34 results
  in a value of 48. The scaling factors are therefore arranged from 5 to 70 to take into account the worst
  possible scaling factors required. If a later main board is used with a substitute transistor for T4, then
  the default scaling factor is about 17 as this transistor's threshold voltage is higher, and thus the AGC
  voltage is lower for S9 and S9+40dB.
*/
void svc_4(void)							// Set S-meter calibration factor
	{
	display_s_meter(GRAPHIC);										// Display actual meter, using the graphic meter scale
	lcd_putst(svc_4_msg);
	get_value(&cal.calval.s_mtr_cal, ENCODER_SLOW, S_METER_CAL_MIN, S_METER_CAL_MAX);
	sprintf((char *)lcdpbuff, cal_factor, cal.calval.s_mtr_cal);	// Added cast - 5B4AIY
	}

void svc_5(void)							// Set annunciator tone length
	{
	get_value(&cal.calval.beep_len, ENCODER_NORMAL, BEEP_LEN_MIN, BEEP_LEN_MAX);

	if(cal.calval.beep_len)
		sprintf((char *)lcdpbuff, svc_5_msg1, cal.calval.beep_len);	// Show current setup - Added cast - 5B4AIY
	else
		sprintf((char *)lcdpbuff, svc_5_msg2, on_off[OFF]);
	}

void svc_6(void)							// Set SWR trip level
	{
	set_decade_value(&cal.calval.swr_trip, MIN_SWR_ALARM, MAX_SWR_ALARM);
	sprintf((char *)lcdpbuff, svc_6_msg, ((double)cal.calval.swr_trip) / 100.0);	// Show current setup - Added cast - 5B4AIY
	}

void svc_7(void)							// Adjust SWR Alarm Delay time
	{
	get_value(&cal.calval.swr_delay, ENCODER_SLOW, MIN_SWR_DLY, MAX_SWR_DLY);

	if(cal.calval.swr_delay)
		sprintf((char *)lcdpbuff, svc_7_msg1, (double)cal.calval.swr_delay);
	else
		sprintf((char *)lcdpbuff, _16s, svc_7_msg3);
	}

void svc_8(void)							// Enable/Disable the Over-Current Alarm
	{
	set_flag(0x01, 0x0E);
	display_alarm_msg(cal.calval.alarm_flags & 0x01);
	}

void svc_9(void)
	{
	if(cal.calval.alarm_flags & 0x01)		// If Over-Current Alarm is enabled,
		{
		cal.calval.overcurrent_trip = (int)set_range((long)cal.calval.overcurrent_trip, SLOW, (long)cal.calval.id_mult, (OVERCURRENT_LOWER * 2000UL), (OVERCURRENT_UPPER * 2000UL));
		sprintf((char *)lcdpbuff, trip_msg, ((double)cal.calval.overcurrent_trip * (double)cal.calval.id_mult) / 2000000.0, 'A');	// Show current setup - Added cast - 5B4AIY
		}
	else									// Else, skip this page
		cal_page++;
	}

void svc_10(void)							// Enable/Disable the Over-Voltage Alarm
	{
	set_flag(0x04, 0x0B);
	display_alarm_msg(cal.calval.alarm_flags & 0x04);
	}

void svc_11(void)
	{
	if(cal.calval.alarm_flags & 0x04)		// If Over-Voltage Alarm is enabled,
		{
		cal.calval.overvoltage_trip = (int)set_range((long)cal.calval.overvoltage_trip, ENCODER_NORMAL, (long)cal.calval.batt_mult, (OVERVOLTAGE_LOWER * 10000UL), (OVERVOLTAGE_UPPER * 10000UL));
		sprintf((char *)lcdpbuff, trip_msg, ((double)cal.calval.overvoltage_trip * (double)cal.calval.batt_mult) / 1000000.0, 'V');		// Show current setup - Added cast - 5B4AIY
		}
	else									// Else, skip this page
		cal_page++;
	}

void svc_12(void)							// Enable/Disable Under-Voltage Alarm
	{
	set_flag(0x08, 0x07);
	display_alarm_msg(cal.calval.alarm_flags & 0x08);
	}

void svc_13(void)
	{
	if(cal.calval.alarm_flags & 0x08)		// If Under-Voltage Alarm is enabled,
		{
		cal.calval.undervoltage_trip = (int)set_range((long)cal.calval.undervoltage_trip, ENCODER_NORMAL, (long)cal.calval.batt_mult, (UNDERVOLTAGE_LOWER * 10000UL), (UNDERVOLTAGE_UPPER * 10000UL));
		sprintf((char *)lcdpbuff, trip_msg, ((double)cal.calval.undervoltage_trip * (double)cal.calval.batt_mult) / 1000000.0, 'V');		// Show current setup - Added cast - 5B4AIY
		}
	else									// Else, skip this page
		cal_page++;
	}

void svc_14(void)							// Select IARU/USA (Extra)/USA (General) Band Limts
	{
	get_value(&cal.calval.Band_Limits, ENCODER_SLOW, IARU_1, USA_GENERAL);
	sprintf(lcdpbuff, _16s, band_limits[cal.calval.Band_Limits]);
	}

void svc_15(void)							// Display load defaults message
	{
	sprintf((char *)lcdpbuff, svc_15_msg);
	}

void (*set_svc[])(void) = {			// Indexed by cal_page
							svc_0,	// 0	Set Master Clock calibration factor
							svc_1,	// 1	Set battery voltage calibration factor
							svc_2,	// 2	Set drain current calibration factor
							svc_3,	// 3	Set power meter calibration factor
							svc_4,	// 4	Set S-meter calibration factor
							svc_5,	// 5	Set annunciator tone length
							svc_6,	// 6	Set SWR trip level
							svc_7,	// 7	Adjust SWR Alarm Delay time
							svc_8,	// 8	Enable/Disable the Over-Current Alarm
							svc_9,	// 9	Set Over-Current Alarm threshold if enabled
							svc_10,	// 10	Enable/Disable the Over-Voltage Alarm
							svc_11,	// 11	Set Over-Voltage Alarm Threshold if enabled
							svc_12,	// 12	Enable/Disable Under-Voltage Alarm
							svc_13,	// 13	Set Under-Voltage Alarm Threshold if enabled
							svc_14,	// 14	Select IARU/USA (Extra)/USA (General) Band Limits
							svc_15	// 15	Load Defaults
							};

void display_svc_page(void)
	{
	lcd_cmd(LINE1);							// Start at line 1
	lcd_putst(service_prompt[cal_page]);	// Display prompt message
	set_svc[cal_page]();					// Execute adjustment function
	lcd_cmd(LINE2);							// now select line 2,
	lcd_putst(lcdpbuff);					// and display adjustment function's return message
	}

void skip_page(int cal_page, int *page)
	{
	if(cal_page == (*page - 1)) (*page)++;
	if(cal_page == (*page + 1)) (*page)--;
	}

void service(void)
	{
	int i, tmp_atten, tmp_mode_A, page;				// Do not remove the page variable! Strange things happen if you re-use variable i. A.Ryan - 4/DEC/2013
	long tmp_vfo_A;

	tmp_vfo_A = eeprom.defval.vfo[VFO_A];			// Save current VFO-A frequency
	tmp_mode_A = eeprom.defval.mode[VFO_A];			// and mode.
	eeprom.defval.vfo_sel = VFO_A;					// Use VFO-A
	vfo_idx = VFO_A;								// Use VFO-A
	eeprom.defval.vfo[VFO_A] = 14300000;			// and set it to 14.3MHz
	eeprom.defval.mode[VFO_A] = TUNE;				// Force VFO-A to Tune mode
	freq_in_use = 14300000;							// Set 14.3MHz as well,
	dds_load(freq_in_use);							// and load synthesiser.
	set_chgen(0);									// Use original S-Meter scale
	tmp_atten = eeprom.defval.Tune_Attenuation;		// Save the current setting of the Tune mode attenuator
	eeprom.defval.Tune_Attenuation = 0;				// and now force it to 0dB for maximum power.
	cal_page = 0;									// Start at the first page
	clear_lcd();

	if(eeprom.defval.serial_mode == SERIAL_TEST)	// If the serial port protocol is set to TEST
		printf(Display_Next_msg);

	do	{
		long_push = LONG_PUSH;
// PTT 
		if(PTT_IN_IO)				// Set PTT status 1 = Receive, 0 = Transmit
			{
			PTT_IN = RECEIVE;		// Copy I/O bit to status flag
			KEY = OFF;				// Remove carrier
			tune_on = OFF;			// and tune state
			}	
		else
			{
			PTT_IN = TRANSMIT;
			KEY = ON;				// Carrier on
			tune_on = ON;			// and tune state
			}

// Main & PA board SPI bus
		send_packet_to_74hc595();

		if(_DISPLAY && SW5)						// DISPLAY/CONFIG - Select calibration function page
			{
			cal_page++;
			cal_page %= (MAX_SERVICE_PAGES + 1);
			beep(HZ392_01, SHORT_BEEP);

			while(_DISPLAY);					// Wait while DISPLAY/CONFIG button is pressed

			ms_delay(BUTTON_DEBOUNCE);
			reset_encoder();					// Flush encoder
			}

		if(_MODE && SW1)						// Fast Calibration Page Select
			{
			while(_MODE)
				{
				if(long_push == 0)
					{
					beep(HZ466_85, cal.calval.beep_len * 10);	// Long Beep for Fast Calibration Page Switch
					latch = TRUE;
					reset_encoder();			// This fixes the spurious Fast Page Switch bug! Finally!

					while(_MODE);				// Wait for button release
					break;
					}
				}

			ms_delay(BUTTON_DEBOUNCE);

			while(latch)						// While Fast Page Select latch is set...
				{
				page = cal_page;
				page += encoder_get_coarse(ENCODER_MEDIUM);
				max_min(&page, MAX_SERVICE_PAGES, 0);

				if(page == 9 && !(cal.calval.alarm_flags & 0x01))	// If Over-Current Flag is reset,
					skip_page(cal_page, &page);						// then skip page.

				if(page == 11 && !(cal.calval.alarm_flags & 0x04))	// If Over-Voltage Flag is reset,
					skip_page(cal_page, &page);						// then skip page.

				if(page == 13 && !(cal.calval.alarm_flags & 0x08))	// If Under-Voltage Flag is reset,
					skip_page(cal_page, &page);						// then skip page.

				cal_page = page;
				display_svc_page();
				warning_beep();

				if((PWR_SW || _MODE) && latch)
					{
					latch = FALSE;									// Release the page hold latch
					beep(HZ587_31, cal.calval.beep_len);
					}
				}

			while(PWR_SW || _MODE);									// Wait for button release

			reset_encoder();
			}

		display_svc_page();

// Read switches
		if(_VFAST)								// Is FAST/VFAST pressed i.e, save current settings?
			{
			if(cal_page == MAX_SERVICE_PAGES)					// Restore defaults only when cal_page == 15
				{
				beep(HZ466_85, LONG_BEEP);						// Long beep for defaults restore
				i = save_settings(1, 4);						// Restore default factory calibration and configuration settings

				if(eeprom.defval.serial_mode == SERIAL_TEST)	// If serial port protocol is TEST,
					printf(Restore_Defaults, cancelled_restored[i]);	// display appropriate message.
				}
			else												// Normal save operation.
				{
				beep(HZ466_85, SHORT_BEEP);						// Short beep for return to normal operation
				i = save_settings(0, 1);						// Save new calibration settings

				if(eeprom.defval.serial_mode == SERIAL_TEST)	// If serial port mode is TEST, print settings message
					printf(EEPROM_Saved, settings_saved[i]);
				}

			eeprom.defval.vfo[VFO_A] = tmp_vfo_A;				// Now restore VFO-A frequency,
			eeprom.defval.mode[VFO_A] = tmp_mode_A;				// and restore VFO-A mode,
			eeprom.defval.Tune_Attenuation = tmp_atten;			// and restore Tune attenuator setting,
			reset_encoder();									// and finally clear encoder.

			while(_VFAST);										// Wait while FAST/VFAST button is pressed

			return;												// Exit service mode
			}				// End IF(_VFAST)
		} while(TRUE);	// End infinite WHILE loop
	}				// End Function

